/* Main code for Gismu Lookup */
/* $Id: gismu.c,v 1.27 1999/02/02 04:32:01 warner Exp $ */

#include <Pilot.h>
#include "callback.h"

#include "gismu_resources.h"

// ALIGNMENT ALIGNMENT ALIGNMENT!!!! PADDING!
struct GismuRecord {
    Char gismu[5];
    struct {
	unsigned char high;
	unsigned char low;
    } places[5]; // 0 for none, 1 .. 1300ish otherwise
    struct {
	unsigned char high;
	unsigned char low;
    } count;     // 0 for very infrequent
    Char numrafsi;
    Char numrefs;
    Char rafsi[0];
    // from here it gets variable
    // Char rafsi[3][numrafsi];
    // Word refs[numrefs]; // gismu numbers
    // Char description; // null-terminated, two nulls if needed for padding
};

struct GismuCluster {
    unsigned char offsets[4];
    struct GismuRecord firstgismu; // the start of 4 such structs
    // each starting at (char *)cluster + offset[n]
};

struct GismuPlaceCluster {
    Char offsets[4];
    Char firstplace[1]; // the start of 4 null-terminated strings
};

struct RafsiCluster {
    unsigned char rafsi[4][3];
    struct {
	unsigned char high, low;
    } gismu[4];
};

// something I read suggests that it is best if this struct is exactly 256
// bytes long. CONSIDER ALIGNMENT
struct MyPrefs {
    unsigned short current_gismu; // what is currently displayed
    unsigned short selected_gismu; // what is selected in the list
    unsigned short selected_rafsi;
    unsigned char list_mode; // 0 for gismu, 1 for rafsi
    unsigned char incremental_search; // 0 for don't, 1 for do
    unsigned char entry[5];
    unsigned char pad[243];
};

static struct MyPrefs prefs;

#define gismuAppCreator 'Gsmu'
#define gismuDBType 'GisD'
#define gismuDBName "GismuDB"
#define gismuPlacesDBType 'GisP'
#define gismuPlacesDBName "GismuPlacesDB"
#define rafsiDBType 'Rfsi'
#define rafsiDBName "RafsiDB"

/* globals */
//static const GISMU_PER_CLUSTER = 4;
//static const PLACES_PER_CLUSTER = 4;
#define GISMU_PER_CLUSTER 4
#define PLACES_PER_CLUSTER 4
#define RAFSI_PER_CLUSTER 4
static DmOpenRef GismuDB, GismuPlacesDB, RafsiDB;
static UInt NumClusters;
static UInt NumRafsiClusters;
static char *gismu_c_revision = "$Revision: 1.27 $";
//static char *gismu_c_date = "$Date: 1999/02/02 04:32:01 $";

/* from PalmPower magazine
   <http://www.palmpower.com/issuesprint/issue199807/utility.html>
*/
static void *
GetObjectPtr(Word objectID)
{
    FormPtr pFrm;
    pFrm = FrmGetActiveForm(); // get pointer to the current form
    return(FrmGetObjectPtr(pFrm, FrmGetObjectIndex(pFrm, objectID)));
}

/* return -1 if one comes before two, 0 if they are the same, 1 if one > two */
static signed int
compare_gismu(char one[5], char two[5])
{
    if (!two[0] && !two[1] && !two[2] && !two[3] && !two[4]) {
	// nil gismu at the end of the list: always "after" any target
	return (-1);
    }
    if (one[0] < two[0])
	return (-1);
    if (one[0] > two[0])
	return (+1);
    if (one[1] < two[1])
	return (-1);
    if (one[1] > two[1])
	return (+1);
    if (one[2] < two[2])
	return (-1);
    if (one[2] > two[2])
	return (+1);
    if (one[3] < two[3])
	return (-1);
    if (one[3] > two[3])
	return (+1);
    if (one[4] < two[4])
	return (-1);
    if (one[4] > two[4])
	return (+1);
    // found it
    return (0);
}

static int
lookup_gismu_subindex(int cluster, char target[5])
{
    // search within a cluster for a given gismu. For now the clusters are
    // fixed to have 4 gismu. Upon entering, we know that gismu[0] was tested
    // and the target is later than it. If the target is later than gismu[3],
    // the target doesn't exist.
    // test gismu[2]. If it matches, return 2. If the target is later than [2],
    // test gismu[3] and return 3 or return none. If the target was earlier
    // than [2], test [1] and return 1 or return none.
    VoidHand r;
    struct GismuCluster *c;
    struct GismuRecord *g;
    signed int dir, found;

    r = DmGetRecord(GismuDB, cluster);
    c = (struct GismuCluster *)MemHandleLock(r);
    g = (struct GismuRecord *) ((char *)c + c->offsets[2]);
    dir = compare_gismu(target, &(g->gismu[0]));
    if (dir == 0)
	found = 2;
    else if (dir < 0) {
	// target is earlier than 2: either 1 or nothing
	g = (struct GismuRecord *) ((char *)c + c->offsets[1]);
	dir = compare_gismu(target, &(g->gismu[0]));
	if (dir < 0)
	    // earlier than 1, later than 0: approx 0
	    found = -1;
	else if (dir > 0)
	    // later than 1, earlier than 2: approx 1
	    found = -2;
	else
	    found = 1;
    } else {
	// target is later than 2: 3 or nothing
	g = (struct GismuRecord *) ((char *)c + c->offsets[3]);
	dir = compare_gismu(target, &(g->gismu[0]));
	if (dir < 0)
	    // later than 2, earlier than 3, approx 2
	    found = -3;
	else if (dir > 0)
	    // later than 3: approx 3
	    found = -4;
	else
	    found = 3;
    }
    MemHandleUnlock(r);
    DmReleaseRecord(GismuDB, cluster, 0);
    return found;
}

static signed int
lookup_gismu(char target[5])
{
    // do a binary search for a particular gismu. Needs to return a record
    // number and an entry within that record. We could save an extra record
    // open/close by returning the opened record but then we would have to
    // return too many things. quick. dirty. Return -1 for "no such gismu", 0
    // .. max if found. The gismu are guaranteed to be in alphabetical order.
    // There are NumGismu of them.
    int here, top, bottom, dir;
    VoidHand r;
    struct GismuCluster *c;
    struct GismuRecord *g;

    here = NumClusters / 2;
    top = NumClusters;
    bottom = -1;
    while(top > bottom+1) {
	/* check gismu[here] */
	r = DmGetRecord(GismuDB, here);
	c = (struct GismuCluster *)MemHandleLock(r);
	g = (struct GismuRecord *)(&(c->firstgismu));
	// the following could be replaced by
	//  dir = StrNCompare(&(target[0]),&(g->gismu[0]), 5)
	// if we wanted to depend upon PalmOS 2.0 or greater
	dir = compare_gismu(&(target[0]), &(g->gismu[0]));
	MemHandleUnlock(r);
	DmReleaseRecord(GismuDB, here, 0);
	if (dir == +1) {
	    bottom = here;
	    here = (here+top)/2;
	} else if (dir == -1) {
	    top = here;
	    here = (bottom+here)/2;
	} else {
	    // found it
	    return 4*here;
	}
    }
    if (top > 0) {
	// need to check the cluster
	signed int subindex = lookup_gismu_subindex(here, target);
	if (subindex >= 0)
	    return 4*here + subindex;
	else
	    return (subindex + 1 - 4*here);
    }
    // didn't find it.
    return -1;
}

/* return -1 if one comes before two, 0 if they are the same, 1 if one > two */
static signed int
compare_rafsi(unsigned char one[3], unsigned char two[3])
{
    if (!two[0] && !two[1] && !two[2]) {
	// nil entries off the end are always "after" every target
	return (-1);
    }
    if (one[0] < two[0])
	return (-1);
    if (one[0] > two[0])
	return (+1);
    if (one[1] < two[1])
	return (-1);
    if (one[1] > two[1])
	return (+1);
    if (one[2] < two[2])
	return (-1);
    if (one[2] > two[2])
	return (+1);
    // found it
    return (0);
}

static int
lookup_rafsi_subindex(int cluster, unsigned char target[3])
{
    // search within a cluster for a given rafsi. For now the clusters are
    // fixed to have 4 gismu. Upon entering, we know that rafsi[0] was tested
    // and the target is later than it. If the target is later than rafsi[3],
    // the target doesn't exist.
    // test rafsi[2]. If it matches, return 2. If the target is later than [2],
    // test rafsi[3] and return 3 or return none. If the target was earlier
    // than [2], test [1] and return 1 or return none.
    VoidHand r;
    struct RafsiCluster *c;
    signed int dir, found;

    r = DmGetRecord(RafsiDB, cluster);
    c = (struct RafsiCluster *)MemHandleLock(r);
    dir = compare_rafsi(target, c->rafsi[2]);
    if (dir == 0)
	found = 2;
    else if (dir < 0) {
	// target is earlier than 2: either 1 or nothing
	dir = compare_rafsi(target, c->rafsi[1]);
	if (dir < 0)
	    // earlier than 1, later than 0: approx 0
	    found = -1;
	else if (dir > 0)
	    // later than 1, earlier than 2: approx 1
	    found = -2;
	else
	    found = 1;
    } else {
	// target is later than 2: 3 or nothing
	dir = compare_rafsi(target, c->rafsi[3]);
	if (dir < 0)
	    // earlier than 3, later than 2: approx 2
	    found = -3;
	else if (dir > 0)
	    // later than 3: approx 3
	    found = -4;
	else
	    found = 3;
    }
    MemHandleUnlock(r);
    DmReleaseRecord(RafsiDB, cluster, 0);
    return found;
}

static signed int
lookup_rafsi(unsigned char target[3])
{
    // do a binary search for a particular rafsi.
    int here, top, bottom, dir;
    VoidHand r;
    struct RafsiCluster *c;

    here = NumRafsiClusters / 2;
    top = NumRafsiClusters;
    bottom = -1;
    while(top > bottom+1) {
	/* check rafsi[here] */
	r = DmGetRecord(RafsiDB, here);
	c = (struct RafsiCluster *)MemHandleLock(r);
	// the following could be replaced by
	//  dir = StrNCompare(&(target[0]),&(c->rafsi[0]), 3)
	// if we wanted to depend upon PalmOS 2.0 or greater
	dir = compare_rafsi(target, c->rafsi[0]);
	MemHandleUnlock(r);
	DmReleaseRecord(RafsiDB, here, 0);
	if (dir == +1) {
	    bottom = here;
	    here = (here+top)/2;
	} else if (dir == -1) {
	    top = here;
	    here = (bottom+here)/2;
	} else {
	    // found it
	    return 4*here;
	}
    }
    if (top > 0) {
	// need to check the cluster
	signed int subindex = lookup_rafsi_subindex(here, target);
	if (subindex >= 0)
	    return 4*here + subindex;
	else
	    return (subindex+1 - 4*here);
    }
    // didn't find it.
    return -1;
}

static int
rafsi_to_gismu(int rafsi_index)
{
    VoidHand r;
    struct RafsiCluster *c;
    int gismu_index, rafsi_subindex;

    r = DmGetRecord(RafsiDB, rafsi_index / RAFSI_PER_CLUSTER);
    c = (struct RafsiCluster *)MemHandleLock(r);
    rafsi_subindex = rafsi_index % RAFSI_PER_CLUSTER;
    if (!c->rafsi[rafsi_subindex][0] && !c->rafsi[rafsi_subindex][1]
	&& !c->rafsi[rafsi_subindex][2]) {
	// one of the trailing null rafsi: not really a gismu
	gismu_index = -1;
    } else {
	gismu_index = (c->gismu[rafsi_subindex].high << 8) |
	    c->gismu[rafsi_subindex].low;
    }
    MemHandleUnlock(r);
    DmReleaseRecord(RafsiDB, rafsi_index / RAFSI_PER_CLUSTER, 0);
    return gismu_index;
}

static void
display_one_place(struct GismuRecord *p, int which, char Xn, int x, int y)
{
    VoidHand r;
    struct GismuPlaceCluster *place;
    Char *p1, *p2;
    int i, j;
    Word num;

    num = (p->places[which].high << 8) + p->places[which].low;

    if (num) {
	r = DmGetRecord(GismuPlacesDB, num / PLACES_PER_CLUSTER);
	if (r) {
	    int offset;
	    place = (struct GismuPlaceCluster *)MemHandleLock(r);
	    offset = place->offsets[num % PLACES_PER_CLUSTER];
	    p1 = &place->firstplace[offset];
	    // need to turn % into x1 (etc), without sprintf
	    p2 = MemPtrNew(StrLen(p1) + 10);
	    if (p2) {
		for(i=0,j=0; i <= StrLen(p1); i++,j++)
		    if (p1[i] == '%') {
			p2[j++] = 'x';
			p2[j] = Xn;
		    } else
			p2[j] = p1[i];
		WinDrawChars(p2, StrLen(p2), x, y);
		MemPtrFree(p2);
	    } else {
		// out of memory
		WinDrawChars(p1, StrLen(p1), x, y);
	    }
	    MemHandleUnlock(r);
	    DmReleaseRecord(GismuPlacesDB, num / PLACES_PER_CLUSTER, 0);
	}
    }
}

//static const int DEF_X = 40;
//static const int DEF_Y = 20;
#define DEF_X 1
#define DEF_Y 29
#define DEF_SIZE_Y 69

// to test limits: use 'karbi': it has x1..x5, and x5 is wide (needs 120 pxls)
// the rectangle (x 40..160  y 20..90) is for gismu lookups
// index == -1 for "not found"
// draw each place structure element on a separate line: 20,+12,+24,+35,+46,+57
// the second element is the english equivalent: make it stand out
// somehow: bold? more spacing?

static void
display_gismu(int index)
{
    // display gismu[index]
    VoidHand r;
    struct GismuCluster *c;
    struct GismuRecord *p;
    RectangleType rect;
    Char *desc;

    prefs.current_gismu = index;

    // erase everything, draw boxes
    RctSetRectangle(&rect, DEF_X, 17, 119, 12);
    WinEraseRectangle(&rect, 0);
    WinDrawRectangleFrame(simpleFrame, &rect);

    RctSetRectangle(&rect, DEF_X, DEF_Y, 119, DEF_SIZE_Y);
    WinEraseRectangle(&rect, 0);
    WinDrawRectangleFrame(simpleFrame, &rect);

    if (index == -1) {
	// write "not found" in definition space
	WinSetClip(&rect);
	WinDrawChars("not found", 9, DEF_X+10, DEF_Y+12);
	WinResetClip();
	return;
    } 

    r = DmGetRecord(GismuDB, index / GISMU_PER_CLUSTER);
    if (!r) {
	// write "bad gismu index" in definition space
	WinSetClip(&rect);
	WinDrawChars("bad gismu index", 15, DEF_X+10, DEF_Y+12);
	WinResetClip();
	return;
    }	
    c = (struct GismuCluster *)MemHandleLock(r);
    p = (struct GismuRecord *) ((char *)c + 
				c->offsets[index % GISMU_PER_CLUSTER]);

    RctSetRectangle(&rect, DEF_X, 17, 119, 12);
    WinSetClip(&rect);

    {
	FontID oldfont = FntGetFont();
	FntSetFont(boldFont);
	// width test: 'mamta' is probably the widest gismu
	WinDrawChars(p->gismu, 5, DEF_X+1, 17);
	FntSetFont(oldfont);
    }

    /* build the rafsi string */
    if (p->numrafsi) {
	int i = 0;
	int j = 0;
	int rafsinum;
	Char *rafsibuf = MemPtrNew(17);
	if (rafsibuf) {
	    for (rafsinum = 0; rafsinum < p->numrafsi; rafsinum++) {
		rafsibuf[i++] = p->rafsi[j++];
		rafsibuf[i++] = p->rafsi[j++];
		if (p->rafsi[j] & 0x80)
		    // high bit set means an apostrophe must be inserted
		    rafsibuf[i++] = '\'';
		rafsibuf[i++] = p->rafsi[j++] & 0x7f;
		rafsibuf[i++] = ' '; rafsibuf[i++] = ' ';
	    }
	    rafsibuf[i] = 0;
	    WinDrawChars(rafsibuf, StrLen(rafsibuf), 
			 DEF_X+40, 17);
	}
	MemPtrFree(rafsibuf);
    }
    WinResetClip();

    RctSetRectangle(&rect, DEF_X, DEF_Y, 119, DEF_SIZE_Y);
    WinSetClip(&rect);

    // x1
    display_one_place(p, 0, '1', DEF_X+1, DEF_Y);

    // description: bold is cool, but too wide for many gismu
    desc = (char *)(&p->rafsi) + 3*p->numrafsi + 2*p->numrefs;
    WinDrawChars(desc, StrLen(desc), DEF_X+5, DEF_Y+12);

    // x2
    display_one_place(p, 1, '2', DEF_X+1, DEF_Y+24);

    // x3
    display_one_place(p, 2, '3', DEF_X+1, DEF_Y+35);

    // x4
    display_one_place(p, 3, '4', DEF_X+1, DEF_Y+46);

    // x5
    display_one_place(p, 4, '5', DEF_X+1, DEF_Y+57);

    MemHandleUnlock(r);
    DmReleaseRecord(GismuDB, index / GISMU_PER_CLUSTER, 0);
    WinResetClip();
}

static void
GismuListSelectItem(int index)
{
    ListPtr list;

    list = GetObjectPtr(GismuList);

    if (index < 0) {
	// approximate position: scroll the list to show item (-index), but
	// don't change the selection
	LstEraseList(list);
	LstSetTopItem(list, -index);
	LstDrawList(list);
    } else {
	LstSetSelection(list, index);
	if (prefs.list_mode)
	    prefs.selected_rafsi = index;
	else
	    prefs.selected_gismu = index;
    }
}


static void
GismuListDrawItem(UInt itemNum,
		  RectanglePtr bounds,
		  CharPtr *itemsText)
{
    VoidHand r;
    struct GismuCluster *c;
    struct GismuRecord *p;
    CALLBACK_PROLOGUE;

    r = DmGetRecord(GismuDB, itemNum / GISMU_PER_CLUSTER);
    c = (struct GismuCluster *)MemHandleLock(r);
    p = (struct GismuRecord *) ((char *)c + 
				c->offsets[itemNum % GISMU_PER_CLUSTER]);
    // assume everything will fit.. will this cause problems if it fails?
    WinDrawChars(&(p->gismu[0]), 5, bounds->topLeft.x, bounds->topLeft.y);
    MemHandleUnlock(r);
    DmReleaseRecord(GismuDB, itemNum / GISMU_PER_CLUSTER, 0);
    //WinDrawChars("test",4,bounds->topLeft.x,bounds->topLeft.y);

    CALLBACK_EPILOGUE;
}

static void
RafsiListDrawItem(UInt itemNum,
		  RectanglePtr bounds,
		  CharPtr *itemsText)
{
    VoidHand r;
    struct RafsiCluster *c;
    Char rafsi[4];
    int len;
    CALLBACK_PROLOGUE;

    r = DmGetRecord(RafsiDB, itemNum / RAFSI_PER_CLUSTER);
    c = (struct RafsiCluster *)MemHandleLock(r);
    rafsi[0] = c->rafsi[itemNum % RAFSI_PER_CLUSTER][0];
    rafsi[1] = c->rafsi[itemNum % RAFSI_PER_CLUSTER][1];
    rafsi[2] = c->rafsi[itemNum % RAFSI_PER_CLUSTER][2];
    if (!rafsi[0] && !rafsi[1] && !rafsi[2]) {
	// trailing off the end of the list
	rafsi[0] = '~'; rafsi[1] = 0; len = 1;
    } else if (rafsi[2] & 0x80) {
	rafsi[3] = rafsi[2] & 0x7f;
	rafsi[2] = '\'';
	len = 4;
    } else {
	len = 3;
    }

    // assume everything will fit.. will this cause problems if it fails?
    WinDrawChars(rafsi, len, bounds->topLeft.x, bounds->topLeft.y);
    MemHandleUnlock(r);
    DmReleaseRecord(RafsiDB, itemNum / RAFSI_PER_CLUSTER, 0);

    CALLBACK_EPILOGUE;
}

static void
set_list_mode(char mode)
{
    ListPtr list;

    list = GetObjectPtr(GismuList);
    prefs.list_mode = mode;
    if (mode) {
	// rafsi mode
	LstSetListChoices(list, NULL, NumRafsiClusters * RAFSI_PER_CLUSTER);
	LstSetDrawFunction(list, RafsiListDrawItem);
	LstSetSelection(list, prefs.selected_rafsi);
	LstMakeItemVisible(list, prefs.selected_rafsi);
	LstDrawList(list);
    } else {
	LstSetListChoices(list, NULL, NumClusters * GISMU_PER_CLUSTER); 
	/* there are 1300ish gismu, each record is a cluster of 4 */
	LstSetDrawFunction(list, GismuListDrawItem);
	LstSetSelection(list, prefs.selected_gismu);
	LstMakeItemVisible(list, prefs.selected_gismu);
	LstDrawList(list);
    }
}

static Boolean 
PrefsHandleEvent (EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;

    CALLBACK_PROLOGUE

    switch (e->eType) {
      case frmOpenEvent:
	frm = FrmGetActiveForm();
	{
	    // set boxes to current prefs
	    CtlSetValue(GetObjectPtr(GismuPrefsIncSearchButton), 
			prefs.incremental_search);
	}
	FrmDrawForm(frm);
	handled = true;
	break;

      case frmCloseEvent:
	// nothing
	break;

      case menuEvent:
	// hmm.. maybe put version info in a separate dialog cued by a menu
	// item from the about box
	break;

      case ctlSelectEvent:
	switch(e->data.ctlSelect.controlID) {
	  case GismuPrefsOKButton:
	      prefs.incremental_search = 
		  CtlGetValue(GetObjectPtr(GismuPrefsIncSearchButton));
	      // fall through to..
	  case GismuPrefsCancelButton:
	    FrmGotoForm(MainForm);
	    handled = true;
	    break;
	}
	break;


      default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

static Boolean 
AboutBoxHandleEvent (EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;

    CALLBACK_PROLOGUE

    switch (e->eType) {
      case frmOpenEvent:
	frm = FrmGetActiveForm();
	{
	    // substitute in version number
	    int len = StrLen(FrmGetLabel(frm, GismuAbout_gismu_c_ID));
	    char *str = MemPtrNew(len);
	    if (str) {
		StrCopy(str, "gismu.c ");
		StrCat(str, gismu_c_revision);
		FrmCopyLabel(frm, GismuAbout_gismu_c_ID, str);
		MemPtrFree(str);
	    }
	}
	FrmDrawForm(frm);
	handled = true;
	break;

      case frmCloseEvent:
	// nothing
	break;

      case menuEvent:
	// hmm.. maybe put version info in a separate dialog cued by a menu
	// item from the about box
	break;

      case ctlSelectEvent:
	switch(e->data.ctlSelect.controlID) {
	  case GismuAboutOKButton :
	    FrmGotoForm(MainForm);
	    handled = true;
	    break;
	}
	break;


      default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

static void
MainFormInit(void)
{
    FormPtr frm;
    ControlPtr button;

    frm = FrmGetActiveForm();
    /* set up the list */
    if (prefs.list_mode) {  // rafsi mode
	// set the radiobuttons to the right mode
	button = GetObjectPtr(GismuListRafsiButton);
	CtlSetValue(button, 1);
    } else { // gismu mode
	button = GetObjectPtr(GismuListGismuButton);
	CtlSetValue(button, 1);
    }
    set_list_mode(prefs.list_mode);

    // add the text from prefs.entry to the entry
    FrmDrawForm(frm);
    // put the cursor in the entry
    FrmSetFocus(frm, FrmGetObjectIndex(frm, GismuEntry));
    // display the gismu we were displaying last time
    display_gismu(prefs.current_gismu);
}

static void
do_lookup_from_entry(void)
{
    Handle t;
    FieldPtr entry_ptr;
    unsigned char *c;
    unsigned char entry[5] = {0,0,0,0,0};
    int i, gismu_index;

    // first, read the text out of the entry field
    entry_ptr = GetObjectPtr(GismuEntry);
    t = FldGetTextHandle(entry_ptr);
    c = MemHandleLock((VoidHand)t);
    //WinDrawChars(c,StrLen(c), 50, 100);
    for(i=0; i < 5 && i < StrLen(c); i++)
	entry[i] = c[i];
    MemHandleUnlock((VoidHand)t);
    
    if (prefs.list_mode) {
	// rafsi mode
	if (StrLen(c) == 4) {
	    entry[2] = entry[3] | 0x80;
	}
	i = lookup_rafsi(entry);
	/* because the word came from the entry rather than the list, we
	   need to update the list to show the chosen word selected */
	GismuListSelectItem(i);
	if (i < 0)
	    gismu_index = -1;
	else
	    gismu_index = rafsi_to_gismu(i);
	display_gismu(gismu_index);
    } else {
	// gismu mode
	gismu_index = lookup_gismu(entry);
	GismuListSelectItem(gismu_index);
	if (gismu_index < 0)
	    gismu_index = -1;
	display_gismu(gismu_index);
    }
}

static void
MainFormShutdown(void)
{
    // copy the text from the entry out into the prefs database

}

static Boolean 
MainFormHandleEvent (EventPtr e)
{
    Boolean handled = false;

    CALLBACK_PROLOGUE

    switch (e->eType) {
      case frmOpenEvent:
	MainFormInit();
	handled = true;
	break;

      case frmCloseEvent:
	MainFormShutdown();
	break;

      case menuEvent:
	MenuEraseStatus(NULL);

	switch(e->data.menu.itemID) {
	  case MainOptionsAbout:
	    FrmGotoForm(GismuAboutForm);
	    break;
	  case MainOptionsPrefs:
	    FrmGotoForm(GismuPrefsForm);
	    break;
	  case EditCut:
	    FldCut(GetObjectPtr(GismuEntry));
	    break;
	  case EditCopy:
	    FldCopy(GetObjectPtr(GismuEntry));
	    break;	
	  case EditPaste:
	    FldPaste(GetObjectPtr(GismuEntry));
	    break;
	  case EditKeyboard:
	    MenuEraseStatus(0);
	    SysKeyboardDialogV10(); // V2.0 issue
	    break;
	  case EditGraffiti:
	    // V2.0 issue
	    break;
	}

    	handled = true;
	break;

      case lstSelectEvent: // an entry in the gismu list was selected
	  {
	      ListPtr list;
	      int index;

	      list = GetObjectPtr(GismuList);
	      index = LstGetSelection(list);
	      if (prefs.list_mode) {
		  // rafsi mode
		  prefs.selected_rafsi = index;
		  index = rafsi_to_gismu(index);
	      } else {
		  prefs.selected_gismu = index;
	      }
	      display_gismu(index);
	      
	      handled = true;
	      break;
	  }

      case ctlSelectEvent:
	switch(e->data.ctlSelect.controlID) {
	  case GismuListGismuButton :
	    set_list_mode(0);
	    handled = true;
	    break;
	  case GismuListRafsiButton :
	    set_list_mode(1);
	    handled = true;
	    break;
	  case GismuLookupButton : 
	    do_lookup_from_entry();
	    handled = true;
	    break;
	}
	break;

      case keyDownEvent:
	  {
	      /* in general don't handle the event (let it pass through to the
		 entry so it will be added). But some will be caught. */
	      unsigned char chr = e->data.keyDown.chr;

	      if (chr == linefeedChr) {
		  do_lookup_from_entry();
		  handled = true;
	      } else if (chr == pageUpChr || chr == pageDownChr) {
		  ListPtr list;
		  enum directions dir;
		  signed int here;

		  if (chr == pageUpChr)
		      dir = up;
		  else
		      dir = down;
		  list = GetObjectPtr(GismuList);
#if 0
		  // the LstScrollList function is only available with PalmOS2
		  LstScrollList(list, dir, 9);
#else
		  // this is pretty much equivalent. The selection is left
		  // alone. The erase/draw pair is needed because the draw
		  // doesn't erase the list first, so the tails of long words
		  // are left behind.
		  here = list->topItem + (dir == up ? -9 : 9);
		  if (here < 0)
		      here = 0;
		  LstEraseList(list);
		  LstSetTopItem(list, here);
		  LstDrawList(list);
#endif
		  handled = true;
	      } else if (chr == nextFieldChr || chr == prevFieldChr) {
		  ListPtr list;
		  enum directions dir;
		  signed int here;
		  EventType event;

		  if (chr == prevFieldChr)
		      dir = up;
		  else
		      dir = down;
		  list = GetObjectPtr(GismuList);
		  here = LstGetSelection(list) + (dir == up ? -1 : 1);
		  if (here < 0)
		      here = 0;
		  LstSetSelection(list, here);

		  MemSet(&event, sizeof(EventType), 0);
		  event.eType = lstSelectEvent;
		  EvtAddEventToQueue(&event);

		  handled = true;
	      } else {
		  /* if enabled, do a kind of incremental search: after adding
		     each character to the entry, scroll the word list so that
		     the current (incomplete) entry would fit somewhere in the
		     portion of the list that is visible.
		     To manage this, if the key just hit would add characters
		     to the entry, send in a fake linefeed after it to trigger
		     a new search. Do this for anything that hasn't already
		     been handled, isn't a hardkey or softkey, and isn't a
		     backspace */
		  if (prefs.incremental_search &&
		      chr <= lastAsciiChr && 
		      chr != backspaceChr) 
		  {
		      EventType event;
		      MemSet(&event, sizeof(EventType), 0);
		      event.eType = keyDownEvent;
		      event.data.keyDown.chr = linefeedChr;
		      EvtAddEventToQueue(&event);
		      // do *not* mark it as handled: we want the key to
		      // get added to the entry
		  }
	      }
	  }
	  break;

      default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

static Boolean 
ApplicationHandleEvent(EventPtr e)
{
    FormPtr frm;
    Word    formId;
    Boolean handled = false;

    if (e->eType == frmLoadEvent) {
	formId = e->data.frmLoad.formID;
	frm = FrmInitForm(formId);
	FrmSetActiveForm(frm);

	switch(formId) {
	  case MainForm:
	    FrmSetEventHandler(frm, MainFormHandleEvent);
	    break;
	  case GismuAboutForm:
	    FrmSetEventHandler(frm, AboutBoxHandleEvent);
	    break;
	  case GismuPrefsForm:
	    FrmSetEventHandler(frm, PrefsHandleEvent);
	    break;
	}
	handled = true;
    }

    return handled;
}

/* Get preferences, open (or create) app database */
static Word 
StartApplication(void)
{
    GismuDB = DmOpenDatabaseByTypeCreator(gismuDBType, 
					  gismuAppCreator, 
					  dmModeReadOnly);
    if (!GismuDB) {
	// signal an error: this database must be constructed externally
	FrmAlert(MissingGismuDBAlert);
	// PalmOS 1.0 will relaunch this app, but hitting the "Applications"
	// button will let them get away.
	return -1;
    }
    GismuPlacesDB = DmOpenDatabaseByTypeCreator(gismuPlacesDBType, 
						gismuAppCreator,
						dmModeReadOnly);
    if (!GismuPlacesDB) {
	FrmAlert(MissingGismuPlacesDBAlert);
	if (GismuDB)
	    DmCloseDatabase(GismuDB);
	return -1;
    }
    NumClusters = DmNumRecords(GismuDB);
    RafsiDB = DmOpenDatabaseByTypeCreator(rafsiDBType, 
					  gismuAppCreator,
					  dmModeReadOnly);
    if (!RafsiDB) {
	FrmAlert(MissingRafsiDBAlert);
	if (GismuDB)
	    DmCloseDatabase(GismuDB);
	if (GismuPlacesDB)
	    DmCloseDatabase(GismuPlacesDB);
	return -1;
    }
    NumRafsiClusters = DmNumRecords(RafsiDB);

    if (!PrefGetAppPreferencesV10(gismuAppCreator, 2, &prefs, sizeof(prefs))) {
	// need to create prefs record
	prefs.current_gismu = 0;
	prefs.selected_gismu = 0;
	prefs.selected_rafsi = 0;
	prefs.list_mode = 0;
	prefs.incremental_search = 0;
	prefs.entry[0] = 0;
	PrefSetAppPreferencesV10(gismuAppCreator, 2, &prefs, sizeof(prefs));
    }
    FrmGotoForm(MainForm);
    return 0;
}

/* Save preferences, close forms, close app database */
static void 
StopApplication(void)
{
    DmCloseDatabase(GismuDB);
    DmCloseDatabase(GismuPlacesDB);
    DmCloseDatabase(RafsiDB);
    PrefSetAppPreferencesV10(gismuAppCreator, 2, &prefs, sizeof(prefs));
    FrmSaveAllForms();
    FrmCloseAllForms();
}

/* The main event loop */
static void 
EventLoop(void)
{
    Word err;
    EventType e;

    do {
	EvtGetEvent(&e, evtWaitForever);
	if (! SysHandleEvent (&e))
	    if (! MenuHandleEvent (NULL, &e, &err))
		if (! ApplicationHandleEvent (&e))
		    FrmDispatchEvent (&e);
    } while (e.eType != appStopEvent);
}

/* Main entry point; it is unlikely you will need to change this except to
   handle other launch command codes */
DWord 
PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
    Word err;

    if (cmd == sysAppLaunchCmdNormalLaunch) {

	err = StartApplication();
	if (err) return err;

	EventLoop();
	StopApplication();

    } else {
	return sysErrParamErr;
    }

    return 0;
}
